package nz.ac.massey.javaecs;
/**
 * Entity Manager
 * Controls adding and removing entities, and registration and unregistration of components to specific entities.
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 */

import java.util.BitSet;
import java.util.Deque;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.ArrayDeque;
import java.util.ArrayList;

/**
 * Manages data from the perspective of the entity.
 * <p>
 * I.e. Controls adding and removing entities, and registration and deregistration of components.
 */
class EntityManager{
    // According to https://stackoverflow.com/questions/12524826/why-should-i-use-deque-over-stack 
    // ArrayDeque is likely faster than a LinkedList, when used in place of one.
    // We can also supply a size to the constructor of ArrayDeque, which avoids resizing the collection
    // at initialisation time (took 1.4s vs 1.8s for 1M)
    private Deque<Entity> unusedEntities; 
    private List<BitSet> entityRegistrations;
    private int maxSize = 1024;

    /**
     * Initialise the EntityManager with the default max size of 1024
     */
    public EntityManager(){
        unusedEntities = new ArrayDeque<>(maxSize);
        entityRegistrations = new ArrayList<>(maxSize); // Init to maxSize to increase performance

        for (int i = 0; i < maxSize; i++) {
            unusedEntities.add(new Entity(i));
            entityRegistrations.add(null); // Annul the entries, to set list size, but keep the unused 
            // entities in an invalid state
        }
    }

    /***
     * Initialise the EntityManager with the provided maximum size
     * @param maxEntities the maximum number of entities to allow
     */
    public EntityManager(int maxEntities){
        maxSize = maxEntities;
        unusedEntities = new ArrayDeque<>(maxSize);
        entityRegistrations = new ArrayList<>(maxEntities);
        
        for (int i = 0; i < maxEntities; i++) {
            unusedEntities.add(new Entity(i));
            entityRegistrations.add(null);
        }
    }

    /**
     * Creates a new entity
     * @return the index of the new entity
     * @throws NoSuchElementException an exception if there are no more unused entities
     */
    protected Entity addEntity() throws NoSuchElementException{
        Entity result = unusedEntities.remove();
        entityRegistrations.set(result.getValue(), new BitSet());
        return result;
    }

    /**
     * Gets the current maximum size
     * @return the value of currentSize
     */
    public Integer getMaxSize(){
        return maxSize;
    }

    /**
     * Gets the BitSet containing the registrations of the entity.
     * @param entity the entity whose BitSet to retrieve
     * @return the BitSet of the provided entity, or a new, empty BitSet if the result was null or out of bounds
     */
    public BitSet getRegistrations(Entity entity){
        try{
            BitSet registrations = entityRegistrations.get(entity.getValue());
            if (registrations != null){
                return registrations;
            }
            else{
                Engine.writeErr("Registrations not initialised for entity: " + entity.getValue() + "; The entity does not exist");
                return new BitSet();
            }
        }
        catch (IndexOutOfBoundsException e){
            Engine.writeErr("Index out of bounds error getting registrations for " + entity.getValue() + "; The entity might not exist");
            return new BitSet(); // Using a blank BitSet will retain data safety (that is, no data will be modified)
        }
    }

    /**
     * Registers the specified component index to the entity
     * @param component the index of the component to register
     * @param entity the entity to register
     * @return true if the operation was successful
     */
    protected boolean registerComponent(int component, Entity entity){
        if (entity.getValue() >= maxSize){
            Engine.writeErr("Attempted to assign a component to non-existent entity: " + entity.getValue());
            return false;
        }
        else if (entityRegistrations.get(entity.getValue()).get(component))
        {
            Engine.writeErr("Entity: " + entity.getValue() + " is already assigned to component " + component);
            return false;
        }
        else{
            entityRegistrations.get(entity.getValue()).set(component);
            return true;
        }
    }

    /**
     * Adds the entity index back into unusedEntities, and sets the registrations to null
     * <p>
     * <b>Does not handle associated data</b> Use the method in ECS to remove entities cleanly
     * @param entity the entity to remove
     */
    protected void removeEntity(Entity entity){
        unusedEntities.add(entity);
        entityRegistrations.set(entity.getValue(), null);
    }

    /**
     * Sets the entity's registrations to the provided BitSet
     * @param entity the entity to set
     * @param registrations the preset registrations
     */
    protected void setRegistrations(Entity entity, BitSet registrations){
        entityRegistrations.set(entity.getValue(), registrations);
    }

    /**
     * Unregisters the specified component from the entity
     * <p>
     * <b>Does not handle component data</b> Use the method in ECS to remove components cleanly
     * @param component the component index to remove
     * @param entity the entity to remove
     * @return true if successful
     */
    protected boolean unregisterComponent(int component, Entity entity){
        try{
            entityRegistrations.get(entity.getValue()).clear(component);
            return true;
        }
        catch (NullPointerException e){
            Engine.writeErr("Entity not initialised");
            return false;
        }
        catch (IndexOutOfBoundsException e)
        {
            return false;
        }
    }

    /**
     * Resizes the currentSize of the entity manager.
     * @param newSize the new maximum size
     * @param systemManager reference to the instanced SystemManager
     * @param componentManager reference to the insanced ComponentManager
     * @return true if the operation succeeded, otherwise false
     */
    public boolean resize(int newSize, SystemManager systemManager, ComponentManager componentManager){
        if (newSize < maxSize - unusedEntities.size()){
            Engine.writeErr("Attempted to resize the maximum entity count to a number smaller than the current assigned entity count.");
            return false;
        }
        else if (newSize == maxSize){
            Engine.writeErr("Attempted to set the newSize to the current size");
            return true;
        }
        else{
            // Every element above the index must be reordered 
            if (newSize < maxSize){
                // Loop through every entity value above our desired index
                // and attempt to remove it from the unused entity queue
                // If that fails, the entity is in use, and must have its
                // data copied to a lower index 
                Deque<Entity> outOfBounds = new ArrayDeque<>(maxSize - newSize);
                for (int i = newSize; i < maxSize; i++) {
                    Entity entityI = Entity.asEntity(i);
                    if (!unusedEntities.remove(entityI)){
                        // Could not remove element, as it didn't exist (already assigned).
                        // must find it and reassign it a new in-bounds value
                        outOfBounds.add(entityI);
                    }
                }
                // Process all out-of-bounds entities, and destroy them as we go
                while(outOfBounds.size() > 0){
                    Entity old = outOfBounds.remove();
                    Entity newPos = addEntity();
                    setRegistrations(newPos, getRegistrations(old));
                    // Invoke an update in the SystemManager
                    systemManager.entityDestroyed(old);
                    systemManager.entityRegistrationsChanged(newPos, getRegistrations(newPos));
                    // Invoke the change in the components
                    componentManager.moveAllComponentData(old, newPos, getRegistrations(old));
                }
                for (int i = newSize; i < maxSize; i++) {
                    // Remove out-of-bounds registration data
                    entityRegistrations.remove(newSize);
                }
            }
            else{
                // Init unassigned values
                for (int i = maxSize; i < newSize; i++) {
                    unusedEntities.add(new Entity(i));
                    entityRegistrations.add(new BitSet());
                }
            }
            // Finally, set the current size
            maxSize = newSize;
            return true;
        }
    }

    public int getNumEntities(){
        return maxSize - unusedEntities.size();
    }
}